' *******************************************************************************************************
' *                                                                                                     *
' *     AiChip_SynthCog_002.spin                                                                        *
' *                                                                                                     *
' *******************************************************************************************************
{

This is an attempt to develop a digital music synthesiser with the same type of abilities as Savage
Innovations's Soundgin(TM). Each version will show the incremental steps towards achieving that.

This is an enhancement to AiChip_SynthCog_001.spin with wave selection, oscillator control added and
the internal ROM Sine Table is used for wave table generation. 

The loopCounter and loopTimer variables are there for diagnostic purposes. Testing shows this version
of code runs a single oscillator in around 560 cycles, around 7uS, and can support 12 oscillators in a
64uS time frame. Currrent capability will degrade as the Cog code becomes more complicated and needs
to perform more tasks. The Cog code has not been optimised, paricularly with regard to hub accesses.

Potential Bug : The OScPulse (PWM) wave may not be correct. After converting the unsigned oscPwm to a
                signed value $00 to $FF ( pot far left to far right ) the pulse should be smaller to
                the left, larger to the right, at $00 the output should be zero, at $FF full on.
}

CON

' .-----------------------------------------------------------------------------------------------------.
' |     Board Configuration                                                                             |
' `-----------------------------------------------------------------------------------------------------'

  _CLKMODE                      = XTAL1 + PLL16x        ' 80MHz System Clock Speed
  _XINFREQ                      = 5_000_000

OBJ

  tv                            : "TV_Text"             ' Debugging only - can be deleted
  
CON

' .-----------------------------------------------------------------------------------------------------.
' |     Hardware Configuration                                                                          |
' `-----------------------------------------------------------------------------------------------------'

  PIN_LEFT                      = 10                    ' Set -Ve if not used
  PIN_RIGHT                     = 11                    ' Set -Ve if not used
                                                                                
  PIN_TV                        = 12                    ' Base TV Pin

' .-----------------------------------------------------------------------------------------------------.
' |     Synthesiser Configuration                                                                       |
' `-----------------------------------------------------------------------------------------------------'

  PLAYBACK_RATE_HZ              = 15625                 ' 15,625 Hz ( 64uS tick rate )

  OSC_COUNT                     = 1                     ' How many oscillators

' .-----------------------------------------------------------------------------------------------------.
' |     Oscillator Parameter Offsets                                                                    |
' `-----------------------------------------------------------------------------------------------------'

  OSC_RESET                     = 0                     ' WR, 1 = Reset wave ramp, 0 = Normal
  OSC_WAVRMP                    = 1                     ' RD  32-bit unsigned
  OSC_CTRL                      = 2                     ' WR  8-bit unsigned
  OSC_PWM                       = 3                     ' WR  8-bit unsigned
  OSC_WAVINC                    = 4                     ' WR  32-bit unsigned
  OSC_VOLUME                    = 5                     ' WR  8-bit signed
  OSC_OUTLEV                    = 6                     ' RD  32 bit-unsigned, $80_00_00_00 = 0V
  OSC_OUTFLG                    = 7                     ' RD
  
  OSC_PARAMS                    = 8                     ' WR = Write or Read / RD = Read Only

  OSC_CTRL_BIT_WAVTYP           = %00000111             ' Bit 0-2 Wave Type
  OSC_CTRL_BIT_ABS              = %00001000             ' Bit 3   ABS / Full
  OSC_CTRL_BIT_OVF              = %00010000             ' Bit 4   Ovf / Reset
  OSC_CTRL_BIT_APPLYBEND        = %00100000             ' Bit 5   Apply Bend    <-- Not implemented
  OSC_CTRL_BIT_NOTUSED          = %01000000             ' Bit 6   No Function
  OSC_CTRL_BIT_ON               = %10000000             ' Bit 7   On / Off      <-- Not implemented                             

  OSC_CTRL_WAVTYP_SINE          = %000                  ' Sine
  OSC_CTRL_WAVTYP_TRIANGLE      = %001                  ' Triangle
  OSC_CTRL_WAVTYP_SAW           = %010                  ' Saw      / Ramp Down
  OSC_CTRL_WAVTYP_RAMP          = %011                  ' Ramp     / Ramp Up
  OSC_CTRL_WAVTYP_PULSE         = %100                  ' Pulse    / Pwm
  OSC_CTRL_WAVTYP_NOISE         = %101                  ' Noise                 
  OSC_CTRL_WAVTYP_LEVEL         = %110                  ' Level    / Full On
  OSC_CTRL_WAVTYP_VOCAL         = %111                  ' Vocal                 <-- Not implemented

  OSC_OUTFLG_BIT_OVF            = %00000001             ' 1 = Overflow occured
  
VAR 

' .-----------------------------------------------------------------------------------------------------.
' |     Global Variables for Spin Code                                                                  |
' `-----------------------------------------------------------------------------------------------------'

  long freqHzAtOneStepPerTick                           ' Freq stepping wave table one step per tick
  long cntPerUs                                         ' Number of CNT increments per uS
  
VAR ' 

' .-----------------------------------------------------------------------------------------------------.
' |     Parameters for Cog - These must remain in this order unless the Cog Assembler code is updated   |
' `-----------------------------------------------------------------------------------------------------'

  long synthParams[0]
        
  long pinNumberLeft
  long pinNumberRight
  long timeoutPeriod
  long noiseSource
  long oscCount
  long osc[ OSC_PARAMS * OSC_COUNT ]
  long loopCounter
  long loopTimer
   
' *******************************************************************************************************
' *                                                                                                     *
' *     Main Program Code                                                                               *
' *                                                                                                     *
' *******************************************************************************************************
                                   
PUB SynthCog 

  ' .---------------------------------------------------------------------------------------------------.
  ' |     Configure Hardware                                                                            |
  ' `---------------------------------------------------------------------------------------------------'

  tv.start( PIN_TV )
        
  pinNumberLeft   := PIN_LEFT 
  pinNumberRight := PIN_RIGHT | ( PIN_LEFT == PIN_RIGHT )

  timeoutPeriod := CLKFREQ / PLAYBACK_RATE_HZ

  ' .---------------------------------------------------------------------------------------------------.
  ' |     Set Global Spin Variables                                                                     |
  ' `---------------------------------------------------------------------------------------------------'
  
  freqHzAtOneStepPerTick := CLKFREQ / ( timeoutPeriod * 256 )

  cntPerUs := CLKFREQ / 1_000_000
  
  ' .---------------------------------------------------------------------------------------------------.
  ' |     Initialise the Synthesiser                                                                    |
  ' `---------------------------------------------------------------------------------------------------'

  noiseSource := CNT
   
  oscCount := OSC_COUNT

  ' .---------------------------------------------------------------------------------------------------.
  ' |     Configure the Synthesiser Run-Time                                                            |
  ' `---------------------------------------------------------------------------------------------------'
                                    
  osc[ (OSC_COUNT-1) * OSC_PARAMS + OSC_WAVINC ] := GetWaveIncFromFreq( 1_000 )
  osc[ (OSC_COUNT-1) * OSC_PARAMS + OSC_CTRL   ] := OSC_CTRL_WAVTYP_SINE
  osc[ (OSC_COUNT-1) * OSC_PARAMS + OSC_PWM    ] := UnsignedBytePercent(50)
  osc[ (OSC_COUNT-1) * OSC_PARAMS + OSC_VOLUME ] := SignedBytePercent(100)
  
  ' .---------------------------------------------------------------------------------------------------.
  ' |     Run the Synthesiser Driver                                                                    |
  ' `---------------------------------------------------------------------------------------------------'

  cognew( @SynthDriver, @synthParams )

  repeat
    ? noiseSource
    Report

' *******************************************************************************************************
' *                                                                                                     *
' *     Utility Code                                                                                    *
' *                                                                                                     *
' *******************************************************************************************************
         
PRI Report

    tv.str( String($01,"AiChip_SynthCog_002",$0D) )

    tv.str( String($0D,"Maximum Freq ") )  
    tv.dec( PLAYBACK_RATE_HZ >> 1 )
    tv.str( String("Hz, ") )
    tv.dec( timeoutPeriod * 1000 / ( CLKFREQ / 1_000_000 ) )
    tv.str( String("nS tick") )

    tv.str( String($0D,"One Step Frq ") )   
    tv.dec( freqHzAtOneStepPerTick )
    tv.str( String("Hz") )

    tv.str( String($0D,"Wav Step Inc ") )   
    tv.hex( osc[ (OSC_COUNT-1) * OSC_PARAMS + OSC_WAVINC ], 8 )
    tv.str( String(", ~") )   
    tv.dec( osc[ (OSC_COUNT-1) * OSC_PARAMS + OSC_WAVINC ] / $01_00_00_00 * freqHzAtOneStepPerTick )
    tv.str( String("Hz, ~") )   
    tv.dec( $40_00_00_00 / ( osc[ (OSC_COUNT-1) * OSC_PARAMS + OSC_WAVINC ] >> 2 ) )
    tv.str( String(" step") )   

    tv.str( String($0D,"Loop Counter ") )   
    tv.dec( loopCounter )

    tv.str( String($0D,"Oscillators  ") )
    tv.dec( OSC_COUNT )

    tv.str( String($0D,"Loop Timing  ") )   
    tv.dec( loopTimer )
    tv.str( String(" cycles, ") )
    tv.dec( loopTimer * 1000 / cntPerUs )
    tv.str( String("nS, ~") )
    tv.dec( ( loopTimer * 1000 / cntPerUs + 500 ) / 1000 )
    tv.str( String("uS") )
    
PRI GetWaveIncFromFreq( freq )

  result := $01_00_00_00 / freqHzAtOneStepPerTick * freq

PRI UnsignedBytePercent( n )

  result := n * $FF / 100

PRI SignedBytePercent( n )

  if n => 0
    result := n * $7F / 100
  else
    result := ( (-n) * $7F / 100 ) | $80

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Synthesiser Driver Code                                                                         *
' *                                                                                                     *
' *******************************************************************************************************
                                   
SynthDriver   org       0

' .-----------------------------------------------------------------------------------------------------.
' |     Initialise Hardware                                                                             |
' `-----------------------------------------------------------------------------------------------------'
    
              mov       ptr,PAR                 ' Point to parameters

              call      #SetPin                 ' Configure Left Audio Pin
        IF_NC mov       CTRA,pinNumber

              call      #SetPin                 ' Configure Right Audio Pin
        IF_NC mov       CTRB,pinNumber
                
              rdlong    timeoutAdd,ptr          ' Get the timeout period

              mov       timeout,timeoutAdd      ' Set the first timeout
              add       timeout,CNT

' .-----------------------------------------------------------------------------------------------------.
' |     Main Loop                                                                                       |
' `-----------------------------------------------------------------------------------------------------'

Loop          mov       ptr,PAR                 ' Point past initialisation variables
              add       ptr,#12

' .-----------------------------------------------------------------------------------------------------.
' |     Get Noise Source for this loop ( updated in the Spin wrapper )                                  |
' `-----------------------------------------------------------------------------------------------------'
             
              rdlong    noiseBits,ptr
              add       ptr,#4
              
' .-----------------------------------------------------------------------------------------------------.
' |     Handle Oscillators                                                                              |
' `-----------------------------------------------------------------------------------------------------'

OscRun        rdlong    count,ptr                       ' Get how many Oscillators 

OscLoop       mov       oscOutFlg,#0                    ' No flags to set

              add       ptr,#4                          ' Reset Wave Ramp / Synch Oscillator
              rdlong    acc,ptr WZ
              mov       acc,#0                          ' Clear the Reset flag
              wrlong    acc,ptr

              add       ptr,#(OSC_WAVRMP-OSC_RESET)*4   ' Get current Wave Ramp                  
              rdlong    oscWavRmp,ptr           
        IF_NZ mov       oscWavRmp,#0                    ' Set to zero if we reset / synch'd the Oscillator

              add       ptr,#(OSC_CTRL-OSC_WAVRMP)*4    ' Get control bits
              rdlong    oscCtrl,ptr                             
              add       ptr,#(OSC_WAVINC-OSC_CTRL)*4    ' Point to wave ramp increment
              
              mov       idx,oscCtrl                     ' Select waveform
              and       idx,#%00000111
              add       idx,#OscJumpTable
              movs      OscJump_RegS,idx

              mov       oscOutlev,oscWavRmp             ' Pre-load oscOutLev used later

OscJump_regS  jmp       #OscJumpTable                   ' Jump to a jump to wave format handler
              
OscJumpTable  jmp       #OscSine                        ' 000 Sine
              jmp       #OscTriangle                    ' 001 Triangle
              jmp       #OscSaw                         ' 010 Saw      / Ramp Down
              jmp       #OscRamp                        ' 011 Ramp     / Ramp Up
              jmp       #OscPulse                       ' 100 Pulse    / Pwm
              jmp       #OscNoise                       ' 101 Noise
              jmp       #OscLevel                       ' 110 Level

OscVocal    ' ***                                       ' 111 Vocal
              jmp       #OscGotWave

OscLevel      mov       oscOutLev,k_FFFF_FFFF           ' 110 Level / Full On
              jmp       #OscGotWave

OscNoise      add       oscOutLev,noiseBits             ' 101 Noise
              sub       oscOutLev,CNT
              shl       oscWavRmp,#1 WC, NR             '     Set correct sign bit
              rcr       oscOutLev,#1
              rol       noiseBits,#1                    '     Kick noise bits
              jmp       #OscGotWave

OscPulse      sub       ptr,#(OSC_WAVINC-OSC_PWM)*4     ' 100 Pulse / Pwm - Get PWM setting
              rdlong    oscPwm,ptr
              add       ptr,#(OSC_WAVINC-OSC_PWM)*4

              and       oscPwm,#$80 WZ, NR              '     Convert unsigned to signed
        IF_NZ xor       oscPwm,#$7F
              shr       oscOutlev,#24
              cmp       oscOutlev,oscPwm WC, NR
              mov       oscOutLev,#0                    '     Ramp <  Pwm, set $0000_000
        IF_NC jmp       #OscGotWave                     '     Ramp >= Pwm, fall through to set $FFFF_FFFF

OscSaw        xor       oscOutLev,k_FFFF_FFFF           ' 010 Saw / Ramp Down
              jmp       #OscGotWave

OscTriangle   shl       oscOutLev,#1 WC, NR             ' 001 Triangle
        IF_C  xor       oscOutLev,k_FFFF_FFFF
              jmp       #OscGotWave

OscSine       mov       idx,oscWavRmp                   ' 000 Sine - Index into sine wave table
              shr       idx,#18
              and       idx,k_0000_0FFE                 '     Index = 1110 nnnn nnnn nnn0 / $E000+(n*2)
              and       oscWavRmp,k_4000_0000 WZ, NR    '     Invert index if phase 1 or 3
        IF_NZ xor       idx,k_0000_0FFE
              or        idx,k_0000_E000                 
              rdword    oscOutLev,idx                   '     Read the sine wave
              shl       oscOutLev,#15                   '     Move it up towards msb
              and       oscOutLev,k_0000_8000 WZ, NR    '     Adjust lsb's
        IF_NZ or        oscOutLev,k_0000_7FFF
              and       oscWavRmp,k_8000_0000 WZ, NR    '     If phase 0 or 1
        IF_Z  or        oscOutLev,k_8000_0000           '     Shift upwards above $80_00_00_00, else
        IF_NZ xor       oscOutLev,k_7FFF_FFFF           '     Invert it but below $80_00_00_00

OscRamp                                                 ' 011 - Ramp / Ramp Up - Everything already done !

OscGotWave    rdlong    oscWavInc,ptr                   ' Get Wave Ramp Increment

              add       oscWavRmp,oscWavInc WC          ' Update Wave Ramp, check for overflow

  IF_C        or        oscOutFlg,#OSC_OUTFLG_BIT_OVF   ' Set overflowed flag
  
  IF_C        and       oscCtrl,#OSC_CTRL_BIT_OVF WZ    ' Should we reset ramp on overlflow ?
  IF_C_AND_NZ mov       oscWavRmp,#0                    ' Yes
        
              sub       ptr,#(OSC_WAVINC-OSC_WAVRMP)*4  ' Store new Wave Ramp for next time                  
              wrlong    oscWavRmp,ptr           

              add       ptr,#(OSC_VOLUME-OSC_WAVRMP)*4  ' Get output level adjuster
              rdlong    oscVolume,ptr                   ' 8-bit signed

              mov       lhs,oscOutLev                   ' Prepare operands for Multiply later
              mov       rhs,oscVolume

              and       oscCtrl,#OSC_CTRL_BIT_ABS WZ,NR ' Are we using ABS volume control ?
        IF_NZ jmp       #OscLevAbs                      ' Yes - Move to +Ve or -Ve

OscLevFull    ' if oscOutLev => 0
              '   oscOutLev := ( ( ( oscOutLev ^ $7F_FF_FF_FF ) >> 7 ) * oscLevel ) ^ $7F_FF_FF_FF
              ' else
              '   oscOutLev := ( ( ( oscOutLev & $7F_FF_FF_FF ) >> 7 ) * oscLevel ) | $80_00_00_00

              and       rhs,#$80 WZ, NR                 ' Volume increases away from centre ($80)
        IF_NZ and       rhs,#$7F                        ' $00..$7F ( 7-bit )
             
              shl       lhs,#1 WC, NR
              
        IF_NC xor       lhs,k_7FFF_FFFF
        IF_NC shr       lhs,#7
        IF_NC call      #Multiply
        IF_NC xor       lhs,k_7FFF_FFFF

        IF_C  and       lhs,k_7FFF_FFFF
        IF_C  shr       lhs,#7
        IF_C  call      #Multiply
        IF_C  or        lhs,k_8000_0000

              jmp       #OscLevDone
              
OscLevAbs     shr       lhs,#8

              and       rhs,#$80 WZ, NR

        IF_Z  call      #Multiply
        IF_Z  or        lhs,k_8000_0000
        
        IF_NZ and       rhs,#$7F
        IF_NZ call      #Multiply
        IF_NZ mov       rhs,oscVolume
        IF_NZ and       rhs,#$7F
        IF_NZ shl       rhs,#24
        IF_NZ add       lhs,rhs

OscLevDone    mov       oscOutLev,lhs

              add       ptr,#(OSC_OUTLEV-OSC_VOLUME)*4  ' Update Osc output
              wrlong    oscOutLev,ptr            

              add       ptr,#(OSC_OUTFLG-OSC_OUTLEV)*4  ' Update Osc output flags
              rdlong    acc,ptr
              or        acc,oscOutFlg
              wrlong    acc,ptr

              djnz      count,#OscLoop                  ' Run all Oscillators

' .-----------------------------------------------------------------------------------------------------.
' |     Handle Output Mixers                                                                            |
' `-----------------------------------------------------------------------------------------------------'

MixRun        mov       mixLftLev,oscOutLev             ' Set output levels
              mov       mixRgtLev,oscOutLev

' .-----------------------------------------------------------------------------------------------------.
' |     Update Diagnostics                                                                              |
' `-----------------------------------------------------------------------------------------------------'

              add       loopCtr,#1                      ' Udate loop counter
              add       ptr,#4                  
              wrlong    loopCtr,ptr

              mov       acc,CNT                         ' Update loop timer
              sub       acc,loopCnt
              add       ptr,#4
              wrlong    acc,ptr

' .-----------------------------------------------------------------------------------------------------.
' |     Handle Timing and DAC output                                                                    |
' `-----------------------------------------------------------------------------------------------------'

DacRun        waitcnt   timeout,timeoutAdd              ' Wait for timeout 'tick'

              mov       FRQA,mixLftLev                  ' Update the DAC's
              mov       FRQB,mixRgtLev

              mov       loopCnt,CNT                     ' Restart loop timer
              
              jmp       #Loop                           ' Repeat the process forever

' .-----------------------------------------------------------------------------------------------------.
' |     Multiply, acc := lhs * rhs                                                                      |
' `-----------------------------------------------------------------------------------------------------'

Multiply      call      #PushFlags

              mov       acc,#0

Multiply_Loop shr       rhs,#1 WC, WZ

        IF_C  add       acc,lhs
              shl       lhs,#1
        IF_NZ jmp       #Multiply_Loop

              mov       lhs,acc

              call      #PopFlags
              
Multiply_ret  ret

' .-----------------------------------------------------------------------------------------------------.
' |     Flag Preservation - Save and Restore the C and Z flags                                          |
' `-----------------------------------------------------------------------------------------------------'

PushFlags     muxc      flags,#1                       ' Save C      : ---- ---- ---- ---C
              muxnz     flags,#2                       ' save NZ     : ---- ---- ---- --N-
PushFlags_ret ret

PopFlags      shr       flags,#1 WZ, WC, NR            ' Start with  : ---- ---- ---- --NC
PopFlags_ret  ret                                      ' After shift : ---- ---- ---- ---N (C)
            
' .-----------------------------------------------------------------------------------------------------.
' |     Configure Audio Pins and Associated Chip Registers                                              |
' `-----------------------------------------------------------------------------------------------------'

SetPin        rdlong    pinNumber,ptr                   ' read a pin number
              add       ptr,#4
              
              shl       pinNumber,#1 WC, NR             ' Check if negative ( C := 1 )

        IF_NC mov       pinBitMask,#1                   ' Create a pin bit mask
        IF_NC shl       pinBitMask,pinNumber
        IF_NC or        DIRA,pinBitMask                 ' Set as output ( DIRA[pin] := 1 )
        IF_NC or        pinNumber,ctrInit               ' Determine CTRA/CTRB setting

SetPin_ret    ret

' .-----------------------------------------------------------------------------------------------------.
' |     Pre-Defined Constants and Pre-Initialised Variables                                             |
' `-----------------------------------------------------------------------------------------------------'

k_0000_0FFE   long      $0000_0FFE
k_0000_7FFF   long      $0000_7FFF
k_0000_8000   long      $0000_8000
k_0000_E000   long      $0000_E000
k_00FF_FFFF   long      $00FF_FFFF
k_7FFF_FFFF   long      $7FFF_FFFF
k_4000_0000   long      $4000_0000
k_8000_0000   long      $8000_0000
k_FFFF_FFFF   long      $FFFF_FFFF

loopCtr       long      0                               ' Number of loops executed
flags         long      0                               ' Flag preservation ( Z and C )

' .-----------------------------------------------------------------------------------------------------.
' |     Pre-Defined Constants only used during initialisation                                           |
' `-----------------------------------------------------------------------------------------------------'

ctrInit       long      ( %00110 << 26 ) | ( 0 << 9 )   ' CTRA/CTRB initialisation value for DAC PWM

' .-----------------------------------------------------------------------------------------------------.
' |     Variables only used during initialisation                                                       |
' `-----------------------------------------------------------------------------------------------------'

pinNumber     res       1                               ' Pin number for audio pin
pinBitMask    res       1                               ' Bit mask for audio pin

' .-----------------------------------------------------------------------------------------------------.
' |     Run-Time Variables                                                                              |
' `-----------------------------------------------------------------------------------------------------'

timeout       res       1                               ' Point next timeout should occur
timeoutAdd    res       1                               ' Interval between timeouts

loopCnt       res       1                               ' Number of CNT ticks elapsed during loop
noiseBits     res       1                               ' Noise bits for Nois waveform

ptr           res       1                               ' Ptr to parameters in Hub Memory
count         res       1                               ' General purpose loop counter
idx           res       1                               ' general purpose index / pointer

acc           res       1                               ' Accumulator for multiply
lhs           res       1                               ' Lhs operand for multiply
rhs           res       1                               ' Rhs operand for multiply                      
                           
oscCtrl       res       1                               ' Oscillator control bits                      
oscWavRmp     res       1                               ' Index into oscillator wave table
oscWavInc     res       1                               ' Oscillator wave table index increment
oscPwm        res       1                               ' Oscillator PWM value
oscVolume     res       1                               ' Oscillator volume control
oscOutLev     res       1                               ' Oscillator output
oscOutFlg     res       1                               ' Oscillator output flags

mixLftLev     res       1                               ' Left output level
mixRgtLev     res       1                               ' Right output level

              fit       $1F0

' *******************************************************************************************************
' *                                                                                                     *
' *     End of program code                                                                             *
' *                                                                                                     *
' *******************************************************************************************************