' *******************************************************************************************************
' *                                                                                                     *
' *     AiChip_SynthCog_001.spin                                                                        *
' *                                                                                                     *
' *******************************************************************************************************
{

This is an attempt to develop a digital music synthesiser with the same type of abilities as Savage
Innovations's Soundgin(TM). Each version will show the incremental steps towards achieving that.

Unlike AiChip_Synth_XXX.spin, which was written entirely in Spin, this synthesiser is Cog-based allowing
far greater accuracy and much more processing. In order to develop the AiChip_SynthCog synthesiser, it
will be designed from the ground up and reflect the approach used by SoundGin.

This first step is a single oscillator handler looping at 15,635Hz ( 64uS tick rate ) which will allow
playback of sounds up to 7,817Hz. The 256-step sine wave will be played back at one step per tick for
61Hz, higher frequencies will use multiple steps through the sine wave outputing fewer samples per
cycle.

At this stage we do not care about getting accurate step rates for a specified frequency, this is more
proof of concept that a sine wave can be generated at the right ballpark frequency.

This is the simplest possile demonstration with a single oscillator set to 1kHz outputing directly to
the two DAC's controlled by PWM.

The loopCounter and loopTimer variables are there for diagnostic purposes. Testing shows this version of
code runs a single oscillator in 191 cycles, around 2uS, and can support 39 oscillators in a 64uS time
frame. Currrent capability will degrade as the Cog code becomes more complicated and needs to perform
more tasks. The Cog code has not been optimised, paricularly with regard to hub accesses.

This version uses a hard-coded sine wave table in the Cog itself. Future versions will read the wave
table for ROM.

}

CON

' .-----------------------------------------------------------------------------------------------------.
' |     Board Configuration                                                                             |
' `-----------------------------------------------------------------------------------------------------'

  _CLKMODE                      = XTAL1 + PLL16x        ' 80MHz System Clock Speed
  _XINFREQ                      = 5_000_000

OBJ

  tv                            : "TV_Text"             ' Debugging only - can be deleted
  
CON

' .-----------------------------------------------------------------------------------------------------.
' |     Hardware Configuration                                                                          |
' `-----------------------------------------------------------------------------------------------------'

  PIN_LEFT                      = 10                    ' Set -Ve if not used
  PIN_RIGHT                     = 11                    ' Set -Ve if not used
                                                                                
  PIN_TV                        = 12                    ' Base TV Pin

' .-----------------------------------------------------------------------------------------------------.
' |     ynthesiser Configuration                                                                        |
' `-----------------------------------------------------------------------------------------------------'

  PLAYBACK_RATE_HZ              = 15625                 ' 15,625 Hz ( 64uS tick rate )

  OSC_COUNT                     = 1                     ' How many oscillators

' .-----------------------------------------------------------------------------------------------------.
' |     Parameter Offsets                                                                               |
' `-----------------------------------------------------------------------------------------------------'

  OSC_RESET                     = 0                     ' WR, 1 = Reset wave ramp, 0 = Normal
  OSC_WAVRMP                    = 1                     ' RD  32-bit unsigned
  OSC_WAVINC                    = 2                     ' WR  32-bit unsigned
  OSC_OUTLEV                    = 3                     ' RD  32 bit-unsigned, $80_00_00_00 = 0V

  OSC_PARAMS                    = 4                     ' WR = Write or Read / RD = Read Only

VAR 

' .-----------------------------------------------------------------------------------------------------.
' |     Global Variables for Spin Code                                                                  |
' `-----------------------------------------------------------------------------------------------------'

  long freqHzAtOneStepPerTick                           ' Freq stepping wave table one step per tick
  long cntPerUs                                         ' Number of CNT increments per uS
  
VAR ' 

' .-----------------------------------------------------------------------------------------------------.
' |     Parameters for Cog - These must remain in this order unless the Cog Assembler code is updated   |
' `-----------------------------------------------------------------------------------------------------'

  long synthParams[0]
        
  long pinNumberLeft
  long pinNumberRight
  long timeoutPeriod
  long oscCount
  long osc[ OSC_PARAMS * OSC_COUNT ]
  long loopCounter
  long loopTimer
  
' *******************************************************************************************************
' *                                                                                                     *
' *     Main Program Code                                                                               *
' *                                                                                                     *
' *******************************************************************************************************
                                   
PUB SynthCog

  ' Configure hardware

  tv.start( PIN_TV )
        
  pinNumberLeft   := PIN_LEFT 
  pinNumberRight := PIN_RIGHT | ( PIN_LEFT == PIN_RIGHT )

  timeoutPeriod := CLKFREQ / PLAYBACK_RATE_HZ

  ' Configure gloabl variables
  
  freqHzAtOneStepPerTick := CLKFREQ / ( timeoutPeriod * 256 )

  cntPerUs := CLKFREQ / 1_000_000
  
  ' Configure synthesiser
  
  oscCount := OSC_COUNT

  cognew( @SynthDriver, @synthParams )
                                    
  osc[ (OSC_COUNT-1) * OSC_PARAMS + OSC_WAVINC ] := GetWaveIncFromFreq( 1_000 )
                                 
  repeat
    Report

' *******************************************************************************************************
' *                                                                                                     *
' *     Utility Code                                                                                    *
' *                                                                                                     *
' *******************************************************************************************************
         
PRI Report

    tv.str( String($01,"AiChip_SynthCog_001",$0D) )

    tv.str( String($0D,"Maximum Freq ") )  
    tv.dec( PLAYBACK_RATE_HZ >> 1 )
    tv.str( String("Hz, ") )
    tv.dec( timeoutPeriod * 1000 / ( CLKFREQ / 1_000_000 ) )
    tv.str( String("nS tick") )
 
    tv.str( String($0D,"One Step Frq ") )   
    tv.dec( freqHzAtOneStepPerTick )
    tv.str( String("Hz") )

    tv.str( String($0D,"Wav Step Inc ") )   
    tv.hex( osc[ (OSC_COUNT-1) * OSC_PARAMS + OSC_WAVINC ], 8 )
    tv.str( String(", ~") )   
    tv.dec( osc[ (OSC_COUNT-1) * OSC_PARAMS + OSC_WAVINC ] / $01_00_00_00 * freqHzAtOneStepPerTick )
    tv.str( String("Hz, ~") )   
    tv.dec( $40_00_00_00 / ( osc[ (OSC_COUNT-1) * OSC_PARAMS + OSC_WAVINC ] >> 2 ) )
    tv.str( String(" step") )   

    tv.str( String($0D,"Loop Counter ") )   
    tv.dec( loopCounter )

    tv.str( String($0D,"Oscillators  ") )
    tv.dec( OSC_COUNT )

    tv.str( String($0D,"Loop Timing  ") )   
    tv.dec( loopTimer )
    tv.str( String(" cycles, ") )
    tv.dec( loopTimer * 1000 / cntPerUs )
    tv.str( String("nS, ~") )
    tv.dec( ( loopTimer * 1000 / cntPerUs + 500 ) / 1000 )
    tv.str( String("uS") )
    
PRI GetWaveIncFromFreq( freq )

  result := $01_00_00_00 / freqHzAtOneStepPerTick * freq

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Main Cog Driver Code                                                                            *
' *                                                                                                     *
' *******************************************************************************************************
                                   
SynthDriver   org       0

' .-----------------------------------------------------------------------------------------------------.
' |     Initialise hardware                                                                             |
' `-----------------------------------------------------------------------------------------------------'
    
              mov       ptr,PAR                 ' Point to parameters

              call      #SetPin                 ' Configure Left Audio Pin
        IF_NC mov       CTRA,pinNumber

              call      #SetPin                 ' Configure Right Audio Pin
        IF_NC mov       CTRB,pinNumber
                
              rdlong    timeoutAdd,ptr          ' Get the timeout period

              mov       timeout,timeoutAdd      ' Set the first timeout
              add       timeout,CNT

' .-----------------------------------------------------------------------------------------------------.
' |     Main Cog Loop                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

Loop          mov       ptr,PAR                 ' Point past initialisation variables
              add       ptr,#12

' .-----------------------------------------------------------------------------------------------------.
' |     Handle Oscillators                                                                              |
' `-----------------------------------------------------------------------------------------------------'

OscRun        rdlong    count,ptr               ' Get how many Osc's 

OscLoop       add       ptr,#4                  ' Reset Wave Ramp / Synch Osc
              rdlong    oscRstFlg,ptr WZ
              mov       oscRstFlg,#0            ' Clear the Reset flag
              wrlong    oscRstFlg,ptr

              add       ptr,#4                  ' Get current Wave Ramp
              rdlong    oscWavRmp,ptr
        IF_NZ mov       oscWavRmp,#0

              mov       idx,oscWavRmp           ' Set the wave lookup index
              shr       idx,#24
              add       idx,#SineTable
              movs      OscLoop_RegS,idx               
              nop
OscLoop_RegS  mov       oscOutLev,SineTable     ' Read the wave table entry
              shl       oscOutLev,#24                                      

              add       ptr,#4                  ' Get Wave Ramp Increment
              rdlong    oscWavInc,ptr

              add       oscWavRmp,oscWavInc     ' Update Wave Ramp

              sub       ptr,#4                  ' Store new Wave Ramp for next time
              wrlong    oscWavRmp,ptr

              add       ptr,#8                  ' Update Osc output
              wrlong    oscOutLev,ptr 

              djnz      count,#OscLoop          ' Run all Osc's

' .-----------------------------------------------------------------------------------------------------.
' |     Handle Output Mixers                                                                            |
' `-----------------------------------------------------------------------------------------------------'

MixRun        mov       mixLftLev,oscOutLev     ' Set output levels
              mov       mixRgtLev,oscOutLev

' .-----------------------------------------------------------------------------------------------------.
' |     Update Diagnostics                                                                              |
' `-----------------------------------------------------------------------------------------------------'

              add       loopCtr,#1              ' Udate loop counter
              add       ptr,#4                  
              wrlong    loopCtr,ptr

              mov       idx,CNT                 ' Update loop timer
              sub       idx,loopCnt
              add       ptr,#4
              wrlong    idx,ptr
              
' .-----------------------------------------------------------------------------------------------------.
' |     Handle Timing and DAC output                                                                    |
' `-----------------------------------------------------------------------------------------------------'

DacRun        waitcnt   timeout,timeoutAdd      ' Wait for timeout 'tick'

              mov       FRQA,mixLftLev          ' Update the DAC's
              mov       FRQB,mixRgtLev

              mov       loopCnt,CNT             ' Restart loop timer
              
              jmp       #Loop                   ' Repeat the process forever

' .-----------------------------------------------------------------------------------------------------.
' |     Configure Audio Pins and Associated Chip Registers                                              |
' `-----------------------------------------------------------------------------------------------------'

SetPin        rdlong    pinNumber,ptr           ' read a pin number
              add       ptr,#4
              
              shl       pinNumber,#1 WC, NR     ' Check if negative ( C := 1 )

        IF_NC mov       pinBitMask,#1           ' Create a pin bit mask
        IF_NC shl       pinBitMask,pinNumber
        IF_NC or        DIRA,pinBitMask         ' Set as output ( DIRA[pin] := 1 )
        IF_NC or        pinNumber,ctrInit       ' Determine CTRA/CTRB setting

SetPin_ret    ret

' .-----------------------------------------------------------------------------------------------------.
' |     Pre-Defined Wave Table                                                                          |
' `-----------------------------------------------------------------------------------------------------'

SineTable     long 128,131,134,137,140,143,146,149,152,156,159,162,165,168,171,174
              long 176,179,182,185,188,191,193,196,199,201,204,206,209,211,213,216
              long 218,220,222,224,226,228,230,232,234,236,237,239,240,242,243,245
              long 246,247,248,249,250,251,252,252,253,254,254,255,255,255,255,255
              long 255,255,255,255,255,255,254,254,253,252,252,251,250,249,248,247
              long 246,245,243,242,240,239,237,236,234,232,230,228,226,224,222,220
              long 218,216,213,211,209,206,204,201,199,196,193,191,188,185,182,179
              long 176,174,171,168,165,162,159,156,152,149,146,143,140,137,134,131
              long 128,124,121,118,115,112,109,106,103, 99, 96, 93, 90, 87, 84, 81   
              long  79, 76, 73, 70, 67, 64, 62, 59, 56, 54, 51, 49, 46, 44, 42, 39 
              long  37, 35, 33, 31, 29, 27, 25, 23, 21, 19, 18, 16, 15, 13, 12, 10 
              long   9,  8,  7,  6,  5,  4,  3,  3,  2,  1,  1,  0,  0,  0,  0,  0  
              long   0,  0,  0,  0,  0,  0,  1,  1,  2,  3,  3,  4,  5,  6,  7,  8  
              long   9, 10, 12, 13, 15, 16, 18, 19, 21, 23, 25, 27, 29, 31, 33, 35 
              long  37, 39, 42, 44, 46, 49, 51, 54, 56, 59, 62, 64, 67, 70, 73, 76 
              long  79, 81, 84, 87, 90, 93, 96, 99,103,106,109,112,115,118,121,124

' .-----------------------------------------------------------------------------------------------------.
' |     Pre-Defined Constants and Pre-Initialised Variables                                             |
' `-----------------------------------------------------------------------------------------------------'

loopCtr       long      0

' .-----------------------------------------------------------------------------------------------------.
' |     Pre-Defined Constants only used during initialisation                                           |
' `-----------------------------------------------------------------------------------------------------'

ctrInit       long      ( %00110 << 26 ) | ( 0 << 9 )

' .-----------------------------------------------------------------------------------------------------.
' |     Variables only used during initialisation                                                       |
' `-----------------------------------------------------------------------------------------------------'

pinNumber     res       1
pinBitMask    res       1

' .-----------------------------------------------------------------------------------------------------.
' |     Run-Time Variables                                                                              |
' `-----------------------------------------------------------------------------------------------------'

timeout       res       1
timeoutAdd    res       1

loopCnt       res       1

ptr           res       1
count         res       1
idx           res       1
                           
oscRstFlg     res       1
oscWavInc     res       1
oscWavRmp     res       1
oscOutLev     res       1

mixLftLev     res       1
mixRgtLev     res       1

              fit       $1F0

' *******************************************************************************************************
' *                                                                                                     *
' *     End of program code                                                                             *
' *                                                                                                     *
' *******************************************************************************************************